/*

*/

import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.Callable;
import java.util.concurrent.Future;

/**
   Implement our own "executor service" version of a
   thread pool. An executor service has a queue of
   Callable objects.


*/
public class MyExecutorService<V>
{
   private final int poolSize;
   private final Thread[] threadPool;
   private final BlockingQueue<Callable<V>> taskQueue;
   private final BlockingQueue<MyFuture<V>> resultQueue;

   public MyExecutorService(int poolSize)
   {
      this.poolSize = poolSize;
      threadPool = new Thread[poolSize];
      taskQueue   = new LinkedBlockingQueue<>();
      resultQueue = new LinkedBlockingQueue<>();

      // Create, and start, the worker threads.
      for (int i = 0; i < poolSize; ++i)
      {
         threadPool[i] = new WorkerThread(taskQueue, resultQueue);
         threadPool[i].start(); // Start a worker, which will block on its empty queue.
      }
   }


   /**
      Submit a new task to the task queue.

      This blocks if the task queue is full or if
      some other thread is currently adding a task
      to the queue.
   */
   public Future<V> submit(Callable<V> task)
   {
      taskQueue.add(task);
      MyFuture<V> result = new MyFuture<>();
      resultQueue.add(result);
      return result;
   }


   /**
      Submit a new task to the task queue along
      with a Future. This is a bad idea!

      This blocks if the task queue is full or if
      some other thread is currently adding a task
      to the queue.
   */
   public Future<V> submit(Callable<V> task, MyFuture<V> future)
   {
      taskQueue.add(task);
      resultQueue.add(future);
      return future;
   }


   /**
      Send every worker thread a interrupt message
      and then wait for each worker thread to stop.
   */
   public void shutdown()
   {
      System.out.println("***** Shutting down the thread pool.");
      try
      {
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].interrupt();
         }
         for (int i = 0; i < poolSize; ++i)
         {
            threadPool[i].join();
         }
      }
      catch (InterruptedException e)
      {
      }
   }


   /**
      This inner class implements the worker threads that
      are in the "thread pool". The worker threads share a
      queue of Runnable tasks that they should execute.
      A worker thread executes the run() method in each
      Runnable task that it removes from the task queue.

      Notice how the Runnable objects are NOT used to create
      threads! The Runnable objects are simply convenient
      containers for the code that makes up a task.
   */
   class WorkerThread extends Thread
   {
      private final BlockingQueue<Callable<V>> taskQueue;
      private final BlockingQueue<MyFuture<V>> resultQueue;

      public WorkerThread(BlockingQueue<Callable<V>> taskQueue,
                          BlockingQueue<MyFuture<V>> resultQueue)
      {
         this.taskQueue   = taskQueue; // Shared task queue.
         this.resultQueue = resultQueue;
      }

      @Override
      public void run()
      {
         boolean done = false;
         System.out.println("***** Worker thread " + this.getId() + " has started up.");
         while ( (! done) || (! taskQueue.isEmpty()) )
         {
            try
            {
               // Get a Callable from the task queue.
               final Callable<V> task = taskQueue.take();
               // Execute the Callable's call() method in this thread.
               final V result = task.call();
               // Set this task's Future to completed.
               final MyFuture<V> f = resultQueue.take();
               f.complete(result);

               // Check if the thread pool is supposed to shut down.
               if ( Thread.interrupted() )
               {
                  done = true;
               }
            }
            catch (InterruptedException e)
            {
               done = true;
               //System.out.println("***** Worker thread  " + this.getId() + ", InterruptedException.");
            }
            catch (Exception e)
            {
               System.out.println("call() method failed to compute.");
               e.printStackTrace();
            }
         }
         System.out.println("***** Worker thread " + this.getId() + " has shut down.");
      }
   }
}
